/**************************************************************************\
*								  BUFFER.H								   *
*																		   *
*							Linked buffer classes						   *
*																		   *
*							  by Stewart Gordon							   *
*						  August 1997 - August 1998						   *
\**************************************************************************/
#ifndef __BUFFER_H
#define __BUFFER_H

/*#ifndef __EXCEPT.H
#include <except.h>
#endif*/

#ifndef __STDIO_H
#include <stdio.h>
#endif

#ifndef __MISCIDS_H
#include <miscids.h>
#endif

#define ITERATE(buf)	for((buf).resetiter(); !(buf).endofiter(); (buf)++)
#define ITERATEC(buf)	for(; !(buf).endofiter(); (buf)++)

// class buffer - linked buffer base
template <class N> class buffer {
protected:
	struct node {
		N data;
		node *next;
	} *first;
	node *iterator;

public:
	buffer() : first(NULL), iterator(NULL) {}
	~buffer();
	N& top();
	void pop();
	bool isempty();

	void resetiter() { iterator = first; }
	bool endofiter() { return iterator == NULL; }
	N& operator*() { return iterator->data; }
	void operator++() { iterator = iterator->next; }
	void operator++(int) { iterator = iterator->next; }
};

template <class N>
buffer<N>::~buffer() {
	node *next;
	while(first != NULL) {
		next = first->next;
		delete first;
		first = next;
	}
}

template <class N>
N& buffer<N>::top() {
	N &data = first->data;
	return data;
}

template <class N>
void buffer<N>::pop() {
	node *next = first->next;
	delete first;
	first = next;
}

template <class N>
bool buffer<N>::isempty() {
	return(first == NULL);
}

// class stack - last in first out buffer
template <class N> class stack : public buffer<N> {
	public:
		stack();
		~stack();
		bool push(N& newnode);
};

template <class N> stack<N>::stack() : buffer<N>() {}

template <class N> stack<N>::~stack() {
	buffer<N>::~buffer();
}

template <class N>
bool stack<N>::push(N& newnode) {
	node *next = first;
	try {
		first = new node;
		first->data = newnode;
		first->next = next;
	}
	catch(...) {
		first = next;
		return FALSE;
	}
	return TRUE;
}

// class queue - first in first out
template <class N> class queue : public buffer<N> {
	protected:
		node **last;	// pointer to null pointer at back of queue
	public:
		queue() : buffer<N>() { last = &first; }
		~queue();
		bool add(N& newnode);
};

template <class N> queue<N>::~queue() {
	buffer<N>::~buffer();
}

template <class N>
bool queue<N>::add(N& newnode) {
	node *&next = *last;
	try {
		next = new node;
		//if(last == NULL) first = next;
		next->data = newnode;
		next->next = NULL;
		last = &next->next;
	}
	catch(...) {
		next = NULL;
		return FALSE;
	}
	return TRUE;
}

#endif

