/**************************************************************************\
*								  CUBE.CPP								   *
*																		   *
*				Three-dimensional perspective demonstration				   *
*																		   *
*								 Version 1.4			   				   *
*																		   *
*							  by Stewart Gordon							   *
*							 July 1998-July 1999						   *
>**************************************************************************<
* The current version renders at 35-69 fps on my Pentium II 350 when run   *
* under Windows 98.														   *
\**************************************************************************/
#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <keyid.h>
#include <miscids.h>
#include "3dvector.h"
#include "3dgroup.h"
#define DEFVIEW	0
#define PERCENT(x)		((int)(x * 100))

// GLOBALS
view vw;
const view defview[3] = {
	view(50, 120, -300, 370, 85, 0, 0, 192, 140),
	view(-125, 125, -125, 320, 175, -45, -45, 192, 140),
	view(0, 0, 0, 320, 175, 0, 0, 192, 140)
};

int apage = 0;
scolour facecolour(.75, .75, .75, .25, .25, .25);
polygon left(4), right(4), bottom(4), top(4), front(4), back(4);
group cube(6, &left, &right, &bottom, &top, &front, &back);

enum NavMode { ORIGINAL, FIRST } navmode = ORIGINAL;

void spdtest();
rgb& editcolour(rgb&);
void waitVR();
void waitendVR();

int main(void) {
	// request autodetection
	int gdriver = VGA, gmode = VGAMED, errorcode;

	// initialize graphics mode
	initgraph(&gdriver, &gmode, "");

	// read result of initialization
	errorcode = graphresult();

	if (errorcode != grOk) {	// an error occurred
		printf("Graphics error: %s\n", grapherrormsg(errorcode));
		printf("Press any key to halt:");
		getch();
		exit(1);				// return with error code
	}

	// set current view
	vw = defview[DEFVIEW];
	curview = &vw;

	// define the sides of the cube
	const int size = 100;

	for(int pno = 0; pno < 4; pno++) {
		left[pno].x		= bottom[pno].y	= front[pno].z	= -size;
		right[pno].x	= top[pno].y	= back[pno].z	= size;
		bottom[pno].x	= front[pno].y	= left[pno].z
		  = top[pno].x	= back[pno].y	= right[pno].z
		  = size * (1 - ((pno >> 1) << 1));
		front[pno].x	= left[pno].y	= bottom[pno].z
		  = back[pno].x	= right[pno].y	= top[pno].z
		  = size * (1 - 2 * (pno % 3 == 0));
	}
	front.fillcolour = left.fillcolour = top.fillcolour = 7;
	back.fillcolour = right.fillcolour = bottom.fillcolour = 8;

	front.colour = left.colour = top.colour
	  = back.colour = right.colour = bottom.colour = &facecolour;

	vw.initpal(1, 14);

	// main loop
	int k = ' ';
	while(k != '\33') {
		// render and draw the cube
		cube.render();						// render the cube in memory
		vw.pal->overlap();					// prepare palette for new frame
		setactivepage(apage = 1 - apage);	// toggle active page to reduce
		cleardevice();						//   flicker
		cube.draw();
		waitVR();
		setvisualpage(apage);

		k = toupper(getch());
		switch(k) {
			case NUL:
				k = getch();
				switch(navmode) {
				case ORIGINAL:	// Original navigation mode
					switch(k) {
						case UP:		vw.eye.y -= 5; break;
						case PGUP:		vw.eye.z -= 5; break;
						case LEFT:		vw.eye.x += 5; break;
						case RIGHT:		vw.eye.x -= 5; break;
						case DOWN:		vw.eye.y += 5; break;
						case PGDN:		vw.eye.z += 5; break;
						case CTLEFT:	vw.azim += 30;
						case CTRIGHT:	vw.azim -= 15;
							vw.azim %= 360; break;
						case CTUP:		vw.alt += 15; break;
						case CTDOWN:	vw.alt -= 15; break;
						case F3:		navmode = FIRST;
					}
					break;
				case FIRST:		// First-person navigation mode
					switch(k) {
						case UP:	vw.fwd3(5); break;
						case PGUP:	vw.alt += 15; break;
						case LEFT:	vw.azim += 30;
						case RIGHT:	vw.azim -= 15;
							vw.azim %= 360; break;
						case DOWN:	vw.bwd3(5); break;
						case PGDN:	vw.alt -= 15; break;
						case F2:	navmode = ORIGINAL;
					}
				}
				break;
			case ' ':	// switch between view types
				if(vw.vtype == LIGHTSHADE) {
					vw.vtype = WIREFRAME;
				}
				else {
					vw.vtype++;
				}
				break;

			// 2, 4, 6, 8 move vanishing point on screen
			case '2': vw.vpy += 5; break;
			case '4': vw.vpx -= 5; break;
			case '6': vw.vpx += 5; break;
			case '8': vw.vpy -= 5; break;

			// adjust light colour
			case '+':
				vw.lightcolour += rgb(.1, .1, .1);
				vw.lightcolour.clipmax();
				break;
			case '-':
				vw.lightcolour -= rgb(.1, .1, .1);
				vw.lightcolour.clipmin();
				break;

			case '/': vw.vtype = LIGHTSHADE; editcolour(vw.lightcolour);
				break;
			case '\'': vw.vtype = LIGHTSHADE; editcolour(vw.ambcolour);
				break;
			case ',': vw.vtype = LIGHTSHADE; editcolour(facecolour.diffuse);
				break;
			case '.': vw.vtype = LIGHTSHADE; editcolour(facecolour.specular);
				break;

			case '`': spdtest(); break;		// speed test

			// select predefined view
			case 'A': vw = defview[0]; vw.initpal(1, 14); break;
			case 'B': vw = defview[1]; vw.initpal(1, 14); break;
			case 'C': vw = defview[2]; vw.initpal(1, 14);
		}
	}

	// clean up
	closegraph();
	return 0;
}

// Speed test
void spdtest() {
	// wait for second to change on clock
	time_t starttime = time(NULL) + 1;
	while(time(NULL) != starttime);

	// count frames in a second
	for(int fps = 0; time(NULL) == starttime; fps++) {
		cube.render();
		vw.initpal(1, 14);
		setactivepage(apage = 1 - apage);
		cleardevice();
		cube.draw();
		waitendVR();
		setvisualpage(apage);
	}

	// display result
	char buf[30];

	//gotoxy(1, 1);
	sprintf(buf, "Speed: %d frames per second", fps);
	outtextxy(0, 0, buf);
	while(!kbhit());
}

// Edit a colour
rgb& editcolour(rgb& col) {
	int k = 0;
	rgb tempcol = col;

	while(k != ESC && k != CR) {
		char buf[35];
		vw.initpal(1, 14);
		cube.draw();

		// clear space to display text
		setfillstyle(SOLID_FILL, 0); bar(0, 0, 639, textheight(buf));
		// display RGB values
		sprintf(buf, "Red %3d%%, Green %3d%%, Blue %3d%% ",
		  PERCENT(col.red), PERCENT(col.green), PERCENT(col.blue));
		outtextxy(0, 0, buf);
		// display HSV values
		sprintf(buf, "Hue %3d, Sat %3d%%, Value %3d%% ", (int)col.hue(),
		  PERCENT(col.sat()), PERCENT(col.val()));
		outtextxy(320, 0, buf);

		// get key value
		k = getch();
		switch(k) {
			case ESC: col = tempcol; break;
			case 'R': col.red += .01; break;
			case 'G': col.green += .01; break;
			case 'B': col.blue += .01; break;
			case 'r': col.red -= .01; break;
			case 'g': col.green -= .01; break;
			case 'b': col.blue -= .01; break;
		}

		col.clipmin();
		col.clipmax();
	}

	return col;
}

// Wait for vertical retrace
void waitVR() {
	// Delay return from this routine until start of next vertical retrace

	_asm {
		mov    ah, 08H         // Used for bit comparisons later
		mov    dx, 03DAH       // EGA/VGA i/o register
	Wait1:
		in     al, dx
		test   al, ah
		jnz    Wait1           // Ignore current VR if in progress
	Wait2:
		in     al, dx
		test   al, ah
		jz     Wait2           // Wait for VR to begin
	}
}

// Wait for vertical retrace
void waitendVR() {
	// Delay return from this routine until end of current vertical retrace

	_asm {
		mov    ah, 08H			// Used for bit comparisons later
		mov    dx, 03DAH		// EGA/VGA i/o register
	Wait1:
		in     al, dx
		test   al, ah
		jz     Wait1			// Ignore current gap if in progress
	Wait2:
		in     al, dx
		test   al, ah
		jnz    Wait2			// Wait for VR to end
	}
}

